#!/usr/bin/env python3

# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Visualize experiment progress."""

from argparse import ArgumentParser, Namespace
from typing import Any

from actorch.visualizer.loaders import CSVLoader, TensorBoardLoader
from actorch.visualizer.plotters import MatplotlibPlotter, PlotlyPlotter


__all__ = [
    "Visualize",
]


class Visualize:
    """Visualize experiment progress."""

    _LOADER_TYPES = {
        "csv": CSVLoader,
        "tensorboard": TensorBoardLoader,
    }

    _PLOTTER_TYPES = {
        "matplotlib": MatplotlibPlotter,
        "plotly": PlotlyPlotter,
    }

    def main(self, args: "Namespace") -> "None":
        """Script entry point.

        Parameters
        ----------
        args:
            The command line arguments obtained from an
            `argparse.ArgumentParser` (e.g. the default
            parser returned by method `get_default_parser`).

        """
        input_format = getattr(args, "input_format", None)
        if input_format not in self._LOADER_TYPES:
            raise ValueError(
                f"`input_format` ({input_format}) must be in {list(self._LOADER_TYPES.keys())}"
            )
        backend = getattr(args, "backend", None)
        if backend not in self._PLOTTER_TYPES:
            raise ValueError(
                f"`backend` ({backend}) must be in {list(self._PLOTTER_TYPES.keys())}"
            )
        loader = self._LOADER_TYPES[input_format]()
        plotter = self._PLOTTER_TYPES[backend]()
        del args.input_format, args.backend
        kwargs = vars(args)
        data = loader.load(**kwargs)
        plotter.plot(data, **kwargs)

    @classmethod
    def get_default_parser(cls, **parser_kwargs: "Any") -> "ArgumentParser":
        """Return a default command line argument parser for method `main`.

        Parameters
        ----------
        parser_kwargs:
            The parser initialization keyword arguments.

        Returns
        -------
            The default command line argument parser.

        """
        parser_kwargs.setdefault("description", "Visualize experiment progress")
        parser = ArgumentParser(**parser_kwargs)
        subparsers = parser.add_subparsers(
            title="backend", dest="backend", required=True
        )
        for backend, plotter_cls in cls._PLOTTER_TYPES.items():
            backend_help = f"generate plots via {plotter_cls.__name__.replace('Plotter', '')} backend"
            backend_parser = subparsers.add_parser(backend, help=backend_help)
            backend_subparsers = backend_parser.add_subparsers(
                title="input format", dest="input_format", required=True
            )
            for input_format, loader_cls in cls._LOADER_TYPES.items():
                input_format_help = (
                    f"load {loader_cls.__name__.replace('Loader', '')} progress files"
                )
                backend_subparsers.add_parser(
                    input_format,
                    help=input_format_help,
                    parents=[
                        loader_cls.get_default_parser(),
                        plotter_cls.get_default_parser(),
                    ],
                    conflict_handler="resolve",
                )
        return parser


def _main() -> "None":
    try:
        parser = Visualize.get_default_parser()
        args = parser.parse_args()
        print("------------------- Start ------------------")
        visualize = Visualize()
        visualize.main(args)
        print("------------------- Done -------------------")
    except KeyboardInterrupt:
        print("---- Exiting early (Keyboard Interrupt) ----")


if __name__ == "__main__":
    _main()



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Plotters."""

from actorch.visualizer.plotters.matplotlib_plotter import *
from actorch.visualizer.plotters.plotly_plotter import *
from actorch.visualizer.plotters.plotter import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Concatenated distribution."""

from typing import List, Optional, Sequence

import torch
from torch import Size, Tensor
from torch.distributions import Distribution, constraints, utils
from torch.distributions.constraints import Constraint

from actorch.distributions.constraints import cat
from actorch.distributions.utils import is_discrete
from actorch.registry import register


__all__ = [
    "CatDistribution",
]


@register
class CatDistribution(Distribution):
    """Concatenate a sequence of base distributions with identical
    batch shapes along one of their event dimensions.

    """

    has_enumerate_support = False
    arg_constraints = {}

    def __init__(
        self,
        base_distributions: "Sequence[Distribution]",
        dim: "int" = 0,
        validate_args: "Optional[bool]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        base_distributions:
            The base distributions.
        dim:
            The event dimension along which to concatenate.
        validate_args:
            True to validate the arguments, False otherwise.
            Default to ``__debug__``.

        Raises
        ------
        IndexError
            If `dim` is out of range or not integer.
        ValueError
            If batch shapes of the base distributions are not identical, or the
            corresponding expanded event shapes differ along dimensions other
            than the `dim`-th.

        """
        self.base_dists = base_distributions
        event_ndims = max(len(d.event_shape or [1]) for d in base_distributions)
        self._expanded_event_shapes = [
            torch.Size(list(d.event_shape) + [1] * (event_ndims - len(d.event_shape)))
            for d in base_distributions
        ]
        if dim < -event_ndims or dim >= event_ndims or not float(dim).is_integer():
            raise IndexError(
                f"`dim` ({dim}) must be in the integer interval [-{event_ndims}, {event_ndims})"
            )
        self.dim = dim = int(dim) % event_ndims
        batch_shape = base_distributions[0].batch_shape
        for base_dist in base_distributions[1:]:
            if base_dist.batch_shape != batch_shape:
                raise ValueError(
                    f"Batch shapes of all base distributions "
                    f"({[d.batch_shape for d in base_distributions]}) "
                    f"must be identical"
                )
        event_shape = list(self._expanded_event_shapes[0])
        for expanded_event_shape in self._expanded_event_shapes[1:]:
            if (list(expanded_event_shape[:dim] + expanded_event_shape[dim + 1 :])) != (
                event_shape[:dim] + event_shape[dim + 1 :]
            ):
                raise ValueError(
                    f"Expanded event shapes of all base distributions "
                    f"({self._expanded_event_shapes}) must be identical "
                    f"except for the `dim`-th ({dim}) dimension"
                )
            event_shape[dim] += expanded_event_shape[dim]
        super().__init__(batch_shape, torch.Size(event_shape), validate_args)

    @utils.lazy_property
    def is_discrete(self) -> "bool":
        """Whether the distribution is discrete.

        Returns
        -------
            True if the distribution is discrete, False otherwise.

        """
        return all(is_discrete(d) for d in self.base_dists)

    # override
    @property
    def support(self) -> "Constraint":
        return cat(
            [
                constraints.independent(
                    d.support, len(self.event_shape) - len(d.event_shape)
                )
                for d in self.base_dists
            ],
            dim=self.dim - len(self.event_shape),
            lengths=[shape[self.dim] for shape in self._expanded_event_shapes],
        )

    # override
    @property
    def mean(self) -> "Tensor":
        return self._cat([d.mean for d in self.base_dists])

    # override
    @property
    def variance(self) -> "Tensor":
        return self._cat([d.variance for d in self.base_dists])

    # override
    def sample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        return self._cat([d.sample(sample_shape) for d in self.base_dists])

    # override
    def rsample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        return self._cat([d.rsample(sample_shape) for d in self.base_dists])

    # override
    def log_prob(self, value: "Tensor") -> "Tensor":
        chunks = self._split(value)
        return torch.stack(
            [d.log_prob(c) for d, c in zip(self.base_dists, chunks)]
        ).sum(dim=0)

    # override
    def cdf(self, value: "Tensor") -> "Tensor":
        chunks = self._split(value)
        return torch.stack([d.cdf(c) for d, c in zip(self.base_dists, chunks)]).prod(
            dim=0
        )

    # override
    def has_rsample(self) -> "bool":
        return all(d.has_rsample for d in self.base_dists)

    # override
    def entropy(self) -> "Tensor":
        return torch.stack([d.entropy() for d in self.base_dists]).sum(dim=0)

    def _cat(self, inputs: "Sequence[Tensor]") -> "Tensor":
        inputs = [
            x[(...,) + (None,) * (len(self.event_shape) - len(d.event_shape))]
            for x, d in zip(inputs, self.base_dists)
        ]
        return torch.cat(inputs, dim=self.dim - len(self.event_shape))

    def _split(self, input: "Tensor") -> "List[Tensor]":
        split_sizes = [shape[self.dim] for shape in self._expanded_event_shapes]
        chunks = input.split(split_sizes, dim=self.dim - len(self.event_shape))
        return [
            chunk.reshape(
                *input.shape[: input.ndim - len(self.event_shape)],
                *d.event_shape,
            )
            for chunk, d in zip(chunks, self.base_dists)
        ]

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__}({self.base_dists}, dim: {self.dim})"



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Deterministic agent."""

from typing import Optional, Tuple, Union

import numpy as np
import torch
from gym.spaces import Space
from numpy import ndarray
from torch import device

from actorch.agents.agent import Agent
from actorch.registry import register


__all__ = [
    "DeterministicAgent",
]


@register
class DeterministicAgent(Agent):
    """Agent that returns a deterministic prediction."""

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        is_batched: "bool" = False,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        self._policy_state = None
        super().__init__(
            policy,
            observation_space,
            action_space,
            is_batched,
            device,
        )

    # override
    def _reset(self, mask: "ndarray") -> "None":
        if self._policy_state is not None:
            self._policy_state[mask] = 0.0

    # override
    def _predict(self, flat_observation: "ndarray") -> "Tuple[ndarray, ndarray]":
        # Add temporal axis
        flat_observation = flat_observation[:, None, ...]
        input = torch.as_tensor(flat_observation, device=self.device)
        with torch.no_grad():
            _, self._policy_state = self.policy(input, self._policy_state)
        prediction = self.policy.deterministic_prediction
        flat_action = self.policy.decode(prediction).to("cpu").numpy()
        # Remove temporal axis
        flat_action = flat_action[:, 0, ...]
        log_prob = np.zeros(len(flat_action))
        return flat_action, log_prob



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Transformed distribution."""

from typing import Optional

import torch
from torch import Size, Tensor
from torch.distributions import (
    ComposeTransform,
    Distribution,
    IndependentTransform,
    Transform,
)
from torch.distributions import TransformedDistribution as TorchTransformedDistribution
from torch.distributions import utils
from torch.distributions.constraints import Constraint

from actorch.distributions.transforms import SumTransform
from actorch.distributions.utils import is_affine, is_discrete
from actorch.registry import register


__all__ = [
    "TransformedDistribution",
]


@register
class TransformedDistribution(TorchTransformedDistribution):
    """Extended version of `torch.distributions.TransformedDistribution` that
    implements additional properties and methods (e.g. `mean`,`stddev`, entropy`,
    etc.) and handles discrete base distributions correctly.

    Warnings
    --------
    TODO `torch.distributions.TransformedDistribution`.

    """

    @utils.lazy_property
    def is_discrete(self) -> "bool":
        """Whether the distribution is discrete.

        Returns
        -------
            True if the distribution is discrete, False otherwise.

        """
        if self.reduced_dist:
            return is_discrete(self.reduced_dist)
        return is_discrete(self.base_dist)

    # override
    @property
    def support(self) -> "Constraint":
        if self.reduced_dist:
            return self.reduced_dist.support
        return super().support

    # override
    @property
    def mean(self) -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.mean
        if self._is_affine_or_sum_transform:
            return self._transform_base_mean()
        return super().mean

    # override
    @property
    def variance(self) -> "Tensor":
        return self.stddev ** 2

    # override
    @property
    def stddev(self) -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.stddev
        if self._is_affine_or_sum_transform:
            return self._transform_base_stddev()
        return super().stddev

    # override
    def sample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        if self.reduced_dist:
            return self.reduced_dist.sample(sample_shape)
        return super().sample(sample_shape)

    # override
    def rsample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        if self.reduced_dist:
            return self.reduced_dist.rsample(sample_shape)
        return super().rsample(sample_shape)

    # override
    def log_prob(self, value: "Tensor") -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.log_prob(value)
        if self.is_discrete and self._is_bijective_transform:
            return self.base_dist.log_prob(self._transform.inv(value.float()))
        return super().log_prob(value)

    # override
    def cdf(self, value: "Tensor") -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.cdf(value)
        return super().cdf(value)

    # override
    def icdf(self, value: "Tensor") -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.icdf(value)
        return super().icdf(value)

    # override
    @property
    def has_enumerate_support(self) -> "bool":
        if self.reduced_dist:
            return self.reduced_dist.has_enumerate_support
        return super().has_enumerate_support

    # override
    def enumerate_support(self, expand: "bool" = True) -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.enumerate_support(expand)
        return super().enumerate_support(expand)

    # override
    def entropy(self) -> "Tensor":
        if self.reduced_dist:
            return self.reduced_dist.entropy()
        if not self._is_bijective_transform:
            return super().entropy()
        if self.is_discrete:
            return self.base_dist.entropy()
        if not self._is_affine_transform:
            return super().entropy()
        return self._transform_base_entropy()

    @utils.lazy_property
    def reduced_dist(self) -> "Optional[Distribution]":
        """Reduce the transformed distribution based on the applied transforms.

        Returns
        -------
            The reduced distribution.

        See Also
        --------
        actorch.distributions.registries.reduction_registry.reduce

        """
        # Avoid circular import
        from actorch.distributions.registries.reduction_registry import reduce

        # Contextually check for incompatibilities
        # (e.g. non-maskable batch transforms applied to masked distributions)
        reduced_distribution = reduce(
            self.base_dist,
            self._transform,
            validate_args=self._validate_args,
        )
        if isinstance(reduced_distribution, TransformedDistribution):
            # Avoid recursion
            reduced_distribution.reduced_dist = None
        return reduced_distribution

    @utils.lazy_property
    def _is_bijective_transform(self) -> "bool":
        return self._transform.bijective

    @utils.lazy_property
    def _is_affine_transform(self) -> "bool":
        return is_affine(self._transform)

    @utils.lazy_property
    def _is_affine_or_sum_transform(self) -> "bool":
        def check(transform: "Transform") -> "bool":
            if isinstance(transform, ComposeTransform):
                return all(check(p) for p in transform.parts)
            if isinstance(transform, IndependentTransform):
                return check(transform.base_transform)
            return is_affine(transform) or isinstance(transform, SumTransform)

        # Avoid reference cycle
        # (see https://github.com/pytorch/pytorch/blob/c4196bee9324699fa0533b4e4037a9a9894af88d/torch/nn/parallel/scatter_gather.py#L30)
        try:
            result = check(self._transform)
        finally:
            check = None
        return result

    def _transform_base_mean(self) -> "Tensor":
        return self._transform(self.base_dist.mean)

    def _transform_base_stddev(self) -> "Tensor":
        def apply(transform: "Transform", x: "Tensor") -> "Tensor":
            if isinstance(transform, ComposeTransform):
                for part in transform.parts:
                    x = apply(part, x)
                return x
            if isinstance(transform, IndependentTransform):
                return apply(transform.base_transform, x)
            if is_affine(transform):
                shift = transform(torch.zeros_like(x))
                return transform(x) - shift
            if isinstance(transform, SumTransform):
                return transform(x ** 2).sqrt()
            raise NotImplementedError

        # Avoid reference cycle
        # (see https://github.com/pytorch/pytorch/blob/c4196bee9324699fa0533b4e4037a9a9894af88d/torch/nn/parallel/scatter_gather.py#L30)
        try:
            result = apply(self._transform, self.base_dist.stddev)
        finally:
            apply = None
        return result

    def _transform_base_entropy(self) -> "Tensor":
        # If Y = f(X) where f is a diffeomorphism and X a continuous
        # random variable, it can be shown that:
        # H[Y] = H[X] + E_X[(log o abs o det o J o f)(X)],
        # where J is the Jacobian matrix
        # (see https://en.wikipedia.org/wiki/Differential_entropy#Properties_of_differential_entropy)
        # If the Jacobian matrix is constant then:
        # E_X[(log o abs o det o J o f)(X)] = (log o abs o det o J o f)(c),
        # where c can be any value in the distribution support
        event_dim = len(self.base_dist.event_shape)
        result = self.base_dist.entropy()
        x = self.base_dist.sample()
        for transform in self.transforms:
            y = transform(x)
            event_dim += transform.codomain.event_dim - transform.domain.event_dim
            result += utils._sum_rightmost(
                transform.log_abs_det_jacobian(x, y),
                event_dim - transform.codomain.event_dim,
            )
            x = y
        return result

    @utils.lazy_property
    def _transform(self) -> "Transform":
        if len(self.transforms) > 1:
            return ComposeTransform(self.transforms)
        return self.transforms[0]

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"({self.base_dist}, "
            f"transforms: {self.transforms})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Version according to SemVer versioning system (https://semver.org/)."""


__all__ = [
    "VERSION",
]


_MAJOR = "0"
"""Major version to increment in case of incompatible API changes."""

_MINOR = "0"
"""Minor version to increment in case of backward compatible new functionality."""

_PATCH = "1"
"""Patch version to increment in case of backward compatible bug fixes."""

VERSION = f"{_MAJOR}.{_MINOR}.{_PATCH}"
"""The package version."""



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Agents."""

from actorch.agents.agent import *
from actorch.agents.deterministic_agent import *
from actorch.agents.dltv_bonus_agent import *
from actorch.agents.epsilon_greedy_agent import *
from actorch.agents.gaussian_noise_agent import *
from actorch.agents.ou_noise_agent import *
from actorch.agents.parameter_noise_agent import *
from actorch.agents.stochastic_agent import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Schedules."""

from actorch.schedules.constant_schedule import *
from actorch.schedules.exponential_schedule import *
from actorch.schedules.lambda_schedule import *
from actorch.schedules.linear_schedule import *
from actorch.schedules.schedule import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Matplotlib plotter."""

import os
from argparse import ArgumentParser
from typing import Any, Dict, Tuple

import matplotlib.pyplot as plt
from matplotlib import rc
from numpy import ndarray

from actorch.utils import get_resource
from actorch.visualizer.plotters.plotter import Plotter


__all__ = [
    "MatplotlibPlotter",
]


class MatplotlibPlotter(Plotter):
    """Plotter based on Matplotlib backend."""

    @classmethod
    # override
    def get_default_parser(cls, **parser_kwargs: "Any") -> "ArgumentParser":
        parser = super().get_default_parser(**parser_kwargs)
        parser.add_argument(
            "--figsize",
            nargs=2,
            default=(7.5, 6.0),
            type=float,
            help="figure size",
        )
        parser.add_argument(
            "-f",
            "--format",
            default="png",
            help="output image format",
        )
        parser.add_argument(
            "-u",
            "--usetex",
            action="store_true",
            help="render text with LaTeX",
        )
        parser.add_argument(
            "--style",
            default=get_resource("styles/default-style.mplstyle"),
            help="absolute or relative path to a Matplotlib style file or name of one of Matplotlib built-in styles",
            dest="style_filepath_or_name",
        )
        return parser

    # override
    def _plot_traces(
        self,
        traces: "Dict[str, Tuple[ndarray, ndarray, ndarray]]",
        x_name: "str",
        y_name: "str",
        output_filepath: "str",
        opacity: "float",
        figsize: "Tuple[float, float]" = (7.5, 6.0),
        format: "str" = "png",
        usetex: "bool" = False,
        style_filepath_or_name: "str" = "classic",
        **kwargs: "Any",
    ) -> "None":
        """Plot traces.

        Parameters
        ----------
        traces:
            The traces, i.e. a dict that maps trial names
            to triplets with the following values:
            - the series to plot along the x-axis;
            - the mean (over sampled series) of the series to plot along the y-axis;
            - the shift (over sampled series) from the mean of the series to plot along the y-axis.
        output_filepath:
            The absolute path to the output file.
        x_name:
            The name of the series to plot along the x-axis.
        y_name:
            The name of the series to plot along the y-axis.
        opacity:
            The confidence interval opacity.
        figsize:
            The figure size.
        format:
            The output image format.
        usetex:
            True to render text with LaTeX, False otherwise.
        style_filepath_or_name:
            The absolute or relative path to a Matplotlib style file or the name of one
            of Matplotlib built-in styles
            (see https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html).

        """
        output_filepath = f"{output_filepath}.{format}"
        if os.path.isfile(style_filepath_or_name):
            style_filepath_or_name = os.path.realpath(style_filepath_or_name)
        if usetex:
            x_name = x_name.replace("_", "\_")  # noqa: W605
            y_name = y_name.replace("_", "\_")  # noqa: W605
        with plt.style.context(style_filepath_or_name):
            rc("text", usetex=usetex)
            fig = plt.figure(figsize=figsize)
            for trial_name, (x, mean, shift) in traces.items():
                # Mean
                label = (
                    trial_name.replace("_", "\_")  # noqa: W605
                    if usetex
                    else trial_name
                )
                plt.plot(x, mean, label=label)
                # Confidence interval
                plt.fill_between(
                    x,
                    y1=mean - shift,
                    y2=mean + shift,
                    alpha=opacity,
                )
            plt.legend()
            plt.xlabel(x_name)
            plt.ylabel(y_name)
            fig.tight_layout()
            plt.savefig(output_filepath, bbox_inches="tight")
            plt.close()



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Setup script."""

import os
import re
import subprocess
import sys

from setuptools import Extension, find_packages, setup


_ROOT_DIRPATH = os.path.dirname(os.path.realpath(__file__))

_WINDOWS_DEPENDENCIES = {
    "torch": "https://download.pytorch.org/whl/cpu/torch_stable.html",
}


def _parse_requirements(requirement_filepath):
    requirement_filepath = os.path.realpath(requirement_filepath)
    is_windows = sys.platform in ["cygwin", "win32", "windows"]
    requirements = []
    with open(requirement_filepath, encoding="utf-8") as f:
        for requirement in f:
            # Ignore lines with `-f` flag
            if requirement.split(" ")[0] == "-f":
                continue
            if not is_windows:
                requirements.append(requirement)
                continue
            package_name = re.split("[=|<>!~]", requirement)[0]
            if package_name not in _WINDOWS_DEPENDENCIES:
                requirements.append(requirement)
                continue
            # Windows-specific requirement
            url = _WINDOWS_DEPENDENCIES[package_name]
            return_code = subprocess.call(
                ["pip", "install", requirement, "-f", url, "--no-cache-dir"]
            )
            if return_code != 0:
                raise RuntimeError(f"{requirement} installation failed")
    return requirements


with open(os.path.join(_ROOT_DIRPATH, "actorch", "version.py")) as f:
    tmp = {}
    exec(f.read(), tmp)
    _VERSION = tmp["VERSION"]
    del tmp

with open(os.path.join(_ROOT_DIRPATH, "README.md"), encoding="utf-8") as f:
    _README = f.read()

_REQUIREMENTS = _parse_requirements(
    os.path.join(_ROOT_DIRPATH, "requirements", "requirements.txt")
)

_REQUIREMENTS_DEV = _parse_requirements(
    os.path.join(_ROOT_DIRPATH, "requirements", "requirements-dev.txt")
)

_REQUIREMENTS_TEST = _parse_requirements(
    os.path.join(_ROOT_DIRPATH, "requirements", "requirements-test.txt")
)


setup(
    name="actorch",
    version=_VERSION,
    description="Deep reinforcement learning framework for fast prototyping based on PyTorch",
    long_description=_README,
    long_description_content_type="text/markdown",
    author="Luca Della Libera",
    author_email="luca.dellalib@gmail.com",
    url="https://github.com/lucadellalib/actorch",
    packages=find_packages(),
    ext_modules=[
        Extension(
            "actorch.buffers.utils",
            sources=["actorch/buffers/utils.c"],
        ),
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Environment :: GPU :: NVIDIA CUDA",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    keywords=["Deep reinforcement learning", "PyTorch"],
    platforms=["OS Independent"],
    include_package_data=True,
    install_requires=_REQUIREMENTS,
    entry_points={"console_scripts": ["actorch=actorch.__main__:main"]},
    extras_require={"dev": _REQUIREMENTS_DEV, "test": _REQUIREMENTS_TEST},
    python_requires=">=3.7",
)



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Unflatten action (possibly batched) environment wrapper."""

from typing import Any, Tuple, Union

from gym import ActionWrapper, Env
from numpy import ndarray

from actorch.envs.batched_env import BatchedEnv
from actorch.envs.spaces import Flat
from actorch.envs.utils import Nested
from actorch.envs.wrappers.wrapper import Wrapper


__all__ = [
    "UnflattenAction",
]


class UnflattenAction(Wrapper, ActionWrapper):
    """(Possibly batched) environment wrapper that unflattens the action."""

    def __init__(self, env: "Union[Env, BatchedEnv]") -> "None":
        super().__init__(env)
        is_batched = isinstance(env.unwrapped, BatchedEnv)
        if is_batched:
            self.single_action_space = Flat(env.single_action_space)
        self.action_space = Flat(env.action_space, is_batched)

    # override
    def step(
        self, action: "ndarray", *args: "Any", **kwargs: "Any"
    ) -> "Tuple[Any, Any, Any, Any]":
        return self.env.step(self.action(action), *args, **kwargs)

    # override
    def action(self, action: "ndarray") -> "Nested":
        return self.action_space.unflatten(action)

    # override
    def reverse_action(self, action: "Nested") -> "ndarray":
        return self.action_space.flatten(action)



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""V-trace."""

from typing import Optional, Tuple, Union

from torch import Tensor
from torch.distributions import Distribution

from actorch.algorithms.value_estimation.generalized_estimator import (
    generalized_estimator,
)


__all__ = [
    "vtrace",
]


def vtrace(
    state_values: "Union[Tensor, Distribution]",
    rewards: "Tensor",
    dones: "Tensor",
    mask: "Tensor",
    log_is_weights: "Tensor",
    next_state_values: "Optional[Union[Tensor, Distribution]]" = None,
    discount: "float" = 0.99,
    num_return_steps: "int" = 1,
    trace_decay: "float" = 1.0,
    leakage: "float" = 1.0,
    max_is_weight_trace: "float" = 1.0,
    max_is_weight_delta: "float" = 1.0,
    max_is_weight_advantage: "float" = 1.0,
    standardize_advantage: "bool" = False,
    epsilon: "float" = 1e-6,
) -> "Tuple[Union[Tensor, Distribution], Tensor]":
    """Compute the (distributional) (leaky) V-trace targets, a.k.a. V-trace(n), and the corresponding
    advantages of a trajectory.

    In the following, let `B` denote the batch size and `T` the trajectory maximum length.

    Parameters
    ----------
    state_values:
        The (distributional) state values (`v_t` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
    rewards:
        The rewards (`r_t` in the literature), shape: ``[B, T]``.
    dones:
        The end-of-episode flags, shape: ``[B, T]``.
    mask:
        The boolean tensor indicating which elements (or batch elements
        if distributional) are valid (True) and which are not (False),
        shape: ``[B, T]``.
    log_is_weights:
        The log importance sampling weights, defined as
        ``target_policy.log_prob(action) - behavior_policy.log_prob(action)``,
        where `target_policy` (`pi` in the literature) is the policy being learned about,
        and `behavior_policy` (`mu` in the literature) the one used to generate the data,
        shape: ``[B, T]``.
    next_state_values:
        The (distributional) next state values (`v_t+1` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
        Default to `state_values` from the second timestep on, with a bootstrap value set to 0.
    discount:
        The discount factor (`gamma` in the literature).
    num_return_steps:
        The number of return steps (`n` in the literature).
    trace_decay:
        The trace-decay parameter (`lambda` in the literature).
    leakage:
        The leakage factor (`alpha` in the literature).
    max_is_weight_trace:
        The maximum importance sampling weight for trace computation (`c_bar` in the literature).
    max_is_weight_delta:
        The maximum importance sampling weight for delta computation (`rho_bar` in the literature).
    max_is_weight_advantage:
        The maximum importance sampling weight for advantage computation.
    standardize_advantage:
        True to standardize the advantages, False otherwise.
    epsilon:
        The term added to the denominators to improve numerical stability.

    Returns
    -------
        - The (distributional) (leaky) V-trace targets
          shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``;
        - the corresponding advantages, shape: ``[B, T]``.

    Raises
    ------
    ValueError
        If an invalid argument value is provided.

    References
    ----------
    .. [1] L. Espeholt, H. Soyer, R. Munos, K. Simonyan, V. Mnih, T. Ward, Y. Doron, V. Firoiu,
           T. Harley, I. Dunning, S. Legg, and K. Kavukcuoglu. "IMPALA: Scalable Distributed
           Deep-RL with Importance Weighted Actor-Learner Architectures". In: ICML. 2018.
           URL: https://arxiv.org/abs/1802.01561
    .. [2] T. Zahavy, Z. Xu, V. Veeriah, M. Hessel, J. Oh, H. Van Hasselt, D. Silver, and S. Singh.
           "A Self-Tuning Actor-Critic Algorithm". In: NeurIPS. 2020, pp. 20913-20924.
           URL: https://arxiv.org/abs/2002.12928

    """
    if leakage < 0.0 or leakage > 1.0:
        raise ValueError(f"`leakage` ({discount}) must be in the interval [0, 1]")
    if max_is_weight_trace <= 0.0:
        raise ValueError(
            f"`max_is_weight_trace` ({max_is_weight_trace}) must be in the interval (0, inf)"
        )
    if max_is_weight_delta <= 0.0:
        raise ValueError(
            f"`max_is_weight_delta` ({max_is_weight_delta}) must be in the interval (0, inf)"
        )
    if max_is_weight_advantage <= 0.0:
        raise ValueError(
            f"`max_is_weight_advantage` ({max_is_weight_advantage}) must be in the interval (0, inf)"
        )
    is_weights = log_is_weights.exp()
    trace_weights = (
        leakage * trace_decay * is_weights.clamp(max=max_is_weight_trace)
        + (1 - leakage) * is_weights
    )
    delta_weights = (
        leakage * is_weights.clamp(max=max_is_weight_delta) + (1 - leakage) * is_weights
    )
    advantage_weights = (
        leakage * is_weights.clamp(max=max_is_weight_advantage)
        + (1 - leakage) * is_weights
    )
    return generalized_estimator(
        state_values=state_values,
        rewards=rewards,
        dones=dones,
        mask=mask,
        trace_weights=trace_weights,
        delta_weights=delta_weights,
        advantage_weights=advantage_weights,
        next_state_values=next_state_values,
        discount=discount,
        num_return_steps=num_return_steps,
        trace_decay=trace_decay,
        standardize_advantage=standardize_advantage,
        epsilon=epsilon,
    )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Runner utilities."""

import importlib
import os
import platform
from datetime import datetime
from types import ModuleType
from typing import Any, Dict

import GPUtil
import psutil

from actorch.utils import normalize_byte_size


__all__ = [
    "get_system_info",
    "import_module",
]


def get_system_info() -> "Dict[str, Any]":
    """Return the system hardware and software information.

    Returns
    -------
        The system hardware and software information.

    """
    system_info = {}
    uname = platform.uname()
    system_info["platform"] = {
        "system": uname.system,
        "node_name": uname.node,
        "release": uname.release,
        "version": uname.version,
        "machine": uname.machine,
        "processor": uname.processor,
    }
    boot_timestamp = psutil.boot_time()
    boot_time = datetime.fromtimestamp(boot_timestamp)
    system_info["boot_time"] = boot_time.strftime("%d/%m/%Y %H:%M:%S")
    cpu_freq = psutil.cpu_freq()
    system_info["cpu"] = {
        "physical_cores": psutil.cpu_count(logical=False),
        "total_cores": psutil.cpu_count(logical=True),
        "min_freq": f"{cpu_freq.min:.2f} MHz",
        "max_freq": f"{cpu_freq.max:.2f} MHz",
        "current_freq": f"{cpu_freq.current:.2f} MHz",
        "usage_per_core": {
            f"{i}": f"{p}%"
            for i, p in enumerate(psutil.cpu_percent(percpu=True, interval=1))
        },
        "total_usage": f"{psutil.cpu_percent()}%",
    }
    virtual_memory = psutil.virtual_memory()
    system_info["memory"] = {
        "total": normalize_byte_size(virtual_memory.total),
        "available": normalize_byte_size(virtual_memory.available),
        "used": normalize_byte_size(virtual_memory.used),
        "percentage": f"{virtual_memory.percent}%",
    }
    system_info["gpu"] = {
        f"{gpu.id}": {
            "uuid": gpu.uuid,
            "name": gpu.name,
            "memory": {
                "total": f"{gpu.memoryTotal} MB",
                "available": f"{gpu.memoryFree} MB",
                "used": f"{gpu.memoryUsed} MB",
                "percentage": f"{gpu.load * 100}%",
            },
            "temperature": f"{gpu.temperature} C",
        }
        for gpu in GPUtil.getGPUs()
    }
    return system_info


def import_module(path: "str") -> "ModuleType":
    """Import a Python module at runtime.

    Parameters
    ----------
    path:
        The absolute or relative path to the module.

    Returns
    -------
        The imported module.

    """
    path = os.path.realpath(path)
    module_name = os.path.splitext(os.path.basename(path))[0]
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Sum transform."""

from typing import Any, Tuple

from torch import Tensor
from torch.distributions import constraints
from torch.distributions.constraints import Constraint

from actorch.distributions.transforms.maskable_transform import MaskableTransform
from actorch.registry import register


__all__ = [
    "SumTransform",
]


@register
class SumTransform(MaskableTransform):
    """Sum the input tensor along one of its event dimensions."""

    is_constant_jacobian = False
    """Whether the Jacobian matrix is constant (i.e. the transform is affine)."""

    def __init__(
        self, in_shape: "Tuple[int, ...]", dim: "int" = 0, cache_size: "int" = 0
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shape:
            The input event shape.
        dim:
            The event dimension along which to sum.
        cache_size:
            The cache size. If 0, no caching is done. If 1, the latest
            single value is cached. Only 0 and 1 are supported.

        Raises
        ------
        IndexError
            If `dim` is out of range or not integer.

        """
        if dim < -len(in_shape) or dim >= len(in_shape) or not float(dim).is_integer():
            raise IndexError(
                f"`dim` ({dim}) must be in the integer interval [-{len(in_shape)}, {len(in_shape)})"
            )
        self.in_shape = in_shape
        self.dim = dim = int(dim) % len(in_shape)
        out_shape = list(in_shape)
        del out_shape[dim]
        self._out_shape = tuple(out_shape)
        super().__init__(cache_size)

    # override
    @property
    def domain(self) -> "Constraint":
        return constraints.independent(constraints.real, len(self.in_shape))

    # override
    @property
    def codomain(self) -> "Constraint":
        return constraints.independent(constraints.real, len(self._out_shape))

    # override
    @property
    def transformed_mask(self) -> "Tensor":
        return self.mask.any(dim=self.dim - len(self.in_shape))

    # override
    def with_cache(self, cache_size: "int" = 1) -> "SumTransform":
        if self._cache_size == cache_size:
            return self
        transform = SumTransform(self.in_shape, self.dim, cache_size)
        transform.mask = self.mask
        return transform

    # override
    def _call(self, x: "Tensor") -> "Tensor":
        return x.masked_fill(~self.mask, 0.0).sum(dim=self.dim - len(self.in_shape))

    # override
    def forward_shape(self, shape: "Tuple[int, ...]") -> "Tuple[int, ...]":
        if len(shape) < len(self.in_shape):
            raise ValueError(
                f"The number of input dimensions ({len(shape)}) "
                f"must be at least {len(self.in_shape)}"
            )
        if shape[-len(self.in_shape) :] != self.in_shape:
            raise ValueError(
                f"Input event shape ({shape[-len(self.in_shape):]}) "
                f"must be equal to {self.in_shape}"
            )
        return shape[: -len(self.in_shape)] + self._out_shape

    # override
    def inverse_shape(self, shape: "Tuple[int, ...]") -> "Tuple[int, ...]":
        if len(shape) < len(self._out_shape):
            raise ValueError(
                f"The number of output dimensions ({len(shape)}) "
                f"must be at least {len(self._out_shape)}"
            )
        if shape[len(shape) - len(self._out_shape) :] != self._out_shape:
            raise ValueError(
                f"Output event shape ({shape[len(shape) - len(self._out_shape):]}) "
                f"must be equal to {self._out_shape}"
            )
        return shape[: len(shape) - len(self._out_shape)] + self.in_shape

    def __eq__(self, other: "Any") -> "bool":
        return (
            isinstance(other, SumTransform)
            and (self.in_shape == other.in_shape)
            and (self.dim == other.dim)
            and (self.mask == other.mask).all()
        )

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(in_shape: {self.in_shape}, "
            f"dim: {self.dim}, "
            f"mask: {self.mask if self.mask.numel() == 1 else self.mask.shape})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Batched experience sampler."""

import time
from typing import Any, Callable, Dict, Iterator, Optional, Tuple, Union

import numpy as np
from numpy import ndarray

from actorch.agents import Agent
from actorch.envs import BatchedEnv
from actorch.samplers.sampler import Sampler


__all__ = [
    "BatchedSampler",
]


class BatchedSampler(Sampler):
    """Sampler that samples batched experiences from a batched environment."""

    def __init__(
        self,
        env: "BatchedEnv",
        agent: "Agent",
        callback_fns: "Optional[Dict[str, Callable[..., None]]]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        env:
            The batched environment to sample batched experiences from.
        agent:
            The agent that interacts with the batched environment.
        callback_fns:
            The callback functions, i.e. a dict with the following key-value pairs:
            - "on_episode_start_fn": the function called after each episode start.
                                     It receives as arguments the batched environment,
                                     the agent, the sampling statistics, and the batch
                                     element index of the started episode;
            - "on_episode_step_fn":  the function called after each episode step.
                                     It receives as arguments the batched environment,
                                     the agent, the sampling statistics, the batched
                                     auxiliary diagnostic information, and the batch
                                     element index of the stepped episode;
            - "on_episode_end_fn":   the function called after each episode end.
                                     It receives as arguments the batched environment,
                                     the agent, the sampling statistics, and the batch
                                     element index of the terminated episode.
            Default to ``{}``.

        Raises
        ------
        TypeError
            If `env` is not instance of `BatchedEnv`.

        See Also
        --------
        Sampler.stats

        """
        if not isinstance(env, BatchedEnv):
            raise TypeError(f"`env` ({env}) must be instance of {BatchedEnv}")
        self._batch_size = len(env)
        try:
            super().__init__(env, agent, callback_fns)
        except TypeError:
            pass

    # override
    def reset(self) -> "None":
        super().reset()
        self._length = np.zeros(self._batch_size, dtype=int)
        self._cumreward = np.zeros(self._batch_size)

    # override
    def sample(
        self,
        num_timesteps: "Optional[Union[int, float]]" = None,
        num_episodes: "Optional[Union[int, float]]" = None,
        render: "bool" = False,
    ) -> "Iterator[Tuple[Dict[str, Any], ndarray]]":
        """Sample batched experiences from the batched environment
        for a given number of timesteps or episodes.

        Parameters
        ----------
        num_timesteps:
            The number of timesteps to sample.
            Must be ``None`` if `num_episodes` is provided.
        num_episodes:
            The number of episodes to sample.
            Must be ``None`` if `num_timesteps` is provided.
        render:
            True to render the environment, False otherwise.

        Yields
        ------
            The sampled batched experience, i.e. a dict with the
            following key-value pairs:
            - "observation":      the batched observation;
            - "action":           the batched action;
            - "log_prob":         the batched action log probability;
            - "reward":           the batched reward;
            - "next_observation": the batched next observation;
            - "done":             the batched end-of-episode flag;
            the batched terminal flag (might differ from `done` if
            the environment has a time limit).

        Warnings
        --------
        Since experiences are sampled in batches of size `B`,
        up to ``B - 1`` extra timesteps or episodes might be sampled.

        """
        yield from super().sample(num_timesteps, num_episodes, render)

    # override
    def _sample(
        self,
        num_timesteps: "Union[int, float]",
        num_episodes: "Union[int, float]",
        render: "bool",
    ) -> "Iterator[Tuple[Dict[str, Any], ndarray]]":
        num_sampled_timesteps, num_sampled_episodes = 0, 0
        # Default callbacks
        on_episode_start_fn = self.callback_fns.get("on_episode_start_fn")
        on_episode_step_fn = self.callback_fns.get("on_episode_step_fn")
        on_episode_end_fn = self.callback_fns.get("on_episode_end_fn")
        while (
            num_sampled_timesteps < num_timesteps
            and num_sampled_episodes < num_episodes
        ):
            if self._next_observation is None:
                # First call
                self._next_observation = self.env.reset()
                if render:
                    render_time = time.time()
                    self.env.render()
                    self._stats["render_time_ms"] += (time.time() - render_time) * 1000
                # Callback
                if on_episode_start_fn:
                    for i in range(self._batch_size):
                        on_episode_start_fn(self.env, self.agent, self._stats, i)
            observation = self._next_observation
            inference_time = time.time()
            action, log_prob = self.agent(observation)
            self._stats["inference_time_ms"] += (time.time() - inference_time) * 1000
            step_time = time.time()
            next_observation, reward, terminal, info = self.env.step(action)
            self._stats["step_time_ms"] += (time.time() - step_time) * 1000
            if render:
                render_time = time.time()
                self.env.render()
                self._stats["render_time_ms"] += (time.time() - render_time) * 1000
            self._next_observation = next_observation
            done = terminal
            self._length += 1
            self._cumreward += reward
            num_sampled_timesteps += self._batch_size
            self._stats["num_timesteps"] = num_sampled_timesteps
            # Callback
            if on_episode_step_fn:
                for i in range(self._batch_size):
                    on_episode_step_fn(self.env, self.agent, self._stats, info, i)

            if terminal.any():
                self._next_observation = self.env.reset(terminal)
                if render:
                    render_time = time.time()
                    self.env.render(terminal)
                    self._stats["render_time_ms"] += (time.time() - render_time) * 1000
                self._stats["episode_lengths"] += self._length[terminal].tolist()
                self._stats["episode_cumrewards"] += self._cumreward[terminal].tolist()
                self._length[terminal] = 0
                self._cumreward[terminal] = 0.0
                num_sampled_episodes += terminal.sum()
                self._stats["num_episodes"] = num_sampled_episodes
                self.agent.reset(terminal)
                done = ~np.array(
                    info[i].get("TimeLimit.truncated", not terminal[i])
                    for i in range(self._batch_size)
                )
                # Callback
                if on_episode_end_fn:
                    for i in np.where(terminal)[0]:
                        on_episode_end_fn(self.env, self.agent, self._stats, i)

            experience = {
                "observation": observation,
                "action": action,
                "log_prob": log_prob,
                "reward": reward,
                "next_observation": next_observation,
                "done": done,
            }
            yield experience, terminal



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Plotly plotter."""

import os
from argparse import ArgumentParser
from typing import Any, Dict, Tuple

import plotly
import yaml
from numpy import ndarray
from plotly import graph_objects as go

from actorch.utils import get_resource
from actorch.visualizer.plotters.plotter import Plotter


__all__ = [
    "PlotlyPlotter",
]


class PlotlyPlotter(Plotter):
    """Plotter based on Plotly backend."""

    @classmethod
    # override
    def get_default_parser(cls, **parser_kwargs: "Any") -> "ArgumentParser":
        parser = super().get_default_parser(**parser_kwargs)
        parser.add_argument(
            "--offline",
            action="store_true",
            help="generate self-contained plots that can be displayed offline",
        )
        parser.add_argument(
            "--template",
            default=get_resource("templates/default-template.yml"),
            help="absolute or relative path to a Plotly template file or name of one of Plotly built-in templates",
            dest="template_filepath_or_name",
        )
        return parser

    # override
    def _plot_traces(
        self,
        traces: "Dict[str, Tuple[ndarray, ndarray, ndarray]]",
        x_name: "str",
        y_name: "str",
        output_filepath: "str",
        opacity: "float",
        offline: "bool" = False,
        template_filepath_or_name: "str" = "plotly",
        **kwargs: "Any",
    ) -> "None":
        """Plot traces.

        Parameters
        ----------
        traces:
            The traces, i.e. a dict that maps trial names
            to triplets with the following values:
            - the series to plot along the x-axis;
            - the mean (over sampled series) of the series to plot along the y-axis;
            - the shift (over sampled series) from the mean of the series to plot along the y-axis.
        output_filepath:
            The absolute path to the output file.
        x_name:
            The name of the series to plot along the x-axis.
        y_name:
            The name of the series to plot along the y-axis.
        opacity:
            The confidence interval opacity.
        offline:
            True to generate self-contained plots that can be displayed offline, False otherwise.
        template_filepath_or_name:
            The absolute or relative path to a Plotly template file or the name of one
            of Plotly built-in templates (see https://plotly.com/python/templates/).

        """
        output_filepath = f"{output_filepath}.html"
        template = template_filepath_or_name
        if os.path.isfile(template_filepath_or_name):
            with open(os.path.realpath(template_filepath_or_name)) as f:
                template = go.layout.Template(yaml.safe_load(f))
        fig = go.Figure()
        for i, (trial_name, (x, mean, shift)) in enumerate(traces.items()):
            # Mean
            color = plotly.colors.DEFAULT_PLOTLY_COLORS[
                i % len(plotly.colors.DEFAULT_PLOTLY_COLORS)
            ]
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mean,
                    line={"color": color},
                    legendgroup=i,
                    mode="lines",
                    name=f"{trial_name}",
                )
            )
            # Confidence interval
            color = color.replace("rgb", "rgba").replace(")", f", {opacity})")
            scatter_kwargs = {
                "legendgroup": i,
                "line": {"width": 0},
                "marker": {"color": color},
                "mode": "lines",
                "name": f"{trial_name} CI",
                "showlegend": False,
            }
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mean + shift,
                    **scatter_kwargs,
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mean - shift,
                    fill="tonexty",
                    fillcolor=color,
                    **scatter_kwargs,
                )
            )
        fig.update_layout(
            xaxis={"title": x_name},
            yaxis={"title": y_name},
            legend={"traceorder": "normal"},
            template=template,
        )
        fig.write_html(output_filepath, include_plotlyjs=True if offline else "cdn")



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Kullback-Leibler divergence registry."""

from typing import Union

import torch
from torch import Tensor
from torch.distributions import (
    Distribution,
    Independent,
    LowRankMultivariateNormal,
    MultivariateNormal,
    Normal,
    kl,
    kl_divergence,
    register_kl,
    utils,
)

from actorch.distributions.cat_distribution import CatDistribution
from actorch.distributions.deterministic import Deterministic
from actorch.distributions.finite import Finite
from actorch.distributions.masked_distribution import MaskedDistribution
from actorch.distributions.transformed_distribution import TransformedDistribution
from actorch.distributions.utils import l2_project


__all__ = []


#####################################################################################################
# Deterministic distribution
#####################################################################################################


@register_kl(Deterministic, Distribution)
@register_kl(Deterministic, TransformedDistribution)  # Avoid ambiguities
def _kl_deterministic_distribution(p: "Deterministic", q: "Distribution") -> "Tensor":
    return -q.log_prob(p.value)


#####################################################################################################
# Finite distribution
#####################################################################################################


@register_kl(Finite, Finite)
def _kl_finite_finite(p: "Finite", q: "Finite") -> "Tensor":
    projected_p = Finite(probs=l2_project(p.atoms, p.probs, q.atoms), atoms=q.atoms)
    return kl._kl_categorical_categorical(projected_p, q)


#####################################################################################################
# Independent distribution
#####################################################################################################


# Adapted from:
# https://github.com/pyro-ppl/pyro/blob/1.6.0/pyro/distributions/kl.py#L24
@register_kl(Independent, Independent)
def _kl_independent_independent(p: "Independent", q: "Independent") -> "Tensor":
    # Handle nested independent distributions
    shared_ndims = min(p.reinterpreted_batch_ndims, q.reinterpreted_batch_ndims)
    p_ndims = p.reinterpreted_batch_ndims - shared_ndims
    q_ndims = q.reinterpreted_batch_ndims - shared_ndims
    base_dist_p = Independent(p.base_dist, p_ndims) if p_ndims else p.base_dist
    base_dist_q = Independent(q.base_dist, q_ndims) if q_ndims else q.base_dist
    result = kl_divergence(base_dist_p, base_dist_q)
    if shared_ndims:
        result = utils._sum_rightmost(result, shared_ndims)
    return result


@register_kl(Independent, Distribution)
def _kl_independent_distribution(p: "Independent", q: "Distribution") -> "Tensor":
    if p.event_shape != q.event_shape:
        raise NotImplementedError
    result = kl_divergence(p.base_dist, q)
    return utils._sum_rightmost(result, p.reinterpreted_batch_ndims)


@register_kl(Independent, MultivariateNormal)
@register_kl(Independent, LowRankMultivariateNormal)
def _kl_independent_multivariate_normal(
    p: "Independent", q: "Union[MultivariateNormal, LowRankMultivariateNormal]"
) -> "Tensor":
    if isinstance(p.base_dist, Deterministic) and len(p.event_shape) == 1:
        return -q.log_prob(p.base_dist.value)
    if isinstance(p.base_dist, Normal) and len(p.event_shape) == 1:
        multivariate_normal_p = MultivariateNormal(
            p.base_dist.loc, scale_tril=p.base_dist.scale.diag_embed()
        )
        return kl_divergence(multivariate_normal_p, q)
    return _kl_independent_distribution(p, q)


@register_kl(Distribution, Independent)
def _kl_distribution_independent(p: "Distribution", q: "Independent") -> "Tensor":
    if p.event_shape != q.event_shape:
        raise NotImplementedError
    result = kl_divergence(p, q.base_dist)
    return utils._sum_rightmost(result, q.reinterpreted_batch_ndims)


@register_kl(MultivariateNormal, Independent)
@register_kl(LowRankMultivariateNormal, Independent)
def _kl_multivariate_normal_independent(
    p: "Union[MultivariateNormal, LowRankMultivariateNormal]", q: "Independent"
) -> "Tensor":
    if isinstance(q.base_dist, Normal) and len(q.event_shape) == 1:
        multivariate_normal_q = MultivariateNormal(
            q.base_dist.loc, scale_tril=q.base_dist.scale.diag_embed()
        )
        return kl_divergence(p, multivariate_normal_q)
    return _kl_distribution_independent(p, q)


#####################################################################################################
# Transformed distribution
#####################################################################################################


@register_kl(TransformedDistribution, TransformedDistribution)
def _kl_transformed_transformed(
    p: "TransformedDistribution", q: "TransformedDistribution"
) -> "Tensor":
    reduced_p, reduced_q = p.reduced_dist, q.reduced_dist
    if reduced_p is None and reduced_q is None:
        return kl._kl_transformed_transformed(p, q)
    return kl_divergence(reduced_p or p, reduced_q or q)


@register_kl(TransformedDistribution, Distribution)
@register_kl(TransformedDistribution, Independent)  # Avoid ambiguities
def _kl_transformed_distribution(
    p: "TransformedDistribution", q: "Distribution"
) -> "Tensor":
    reduced_p = p.reduced_dist
    return kl_divergence(reduced_p, q)


@register_kl(Distribution, TransformedDistribution)
@register_kl(Independent, TransformedDistribution)  # Avoid ambiguities
def _kl_distribution_transformed(
    p: "Distribution", q: "TransformedDistribution"
) -> "Tensor":
    reduced_q = q.reduced_dist
    return kl_divergence(p, reduced_q)


#####################################################################################################
# Masked distribution
#####################################################################################################


@register_kl(MaskedDistribution, MaskedDistribution)
def _kl_masked_masked(p: "MaskedDistribution", q: "MaskedDistribution") -> "Tensor":
    result = kl_divergence(p.base_dist, q.base_dist)
    mask = p.mask & q.mask
    return result.masked_fill(~mask, 0.0)


@register_kl(MaskedDistribution, Distribution)
def _kl_masked_distribution(p: "MaskedDistribution", q: "Distribution") -> "Tensor":
    if not p.mask.all():
        raise NotImplementedError
    return kl_divergence(p.base_dist, q)


@register_kl(Distribution, MaskedDistribution)
def _kl_distribution_masked(p: "Distribution", q: "MaskedDistribution") -> "Tensor":
    if not q.mask.all():
        raise NotImplementedError
    return kl_divergence(p, q.base_dist)


#####################################################################################################
# Concatenated distribution
#####################################################################################################


@register_kl(CatDistribution, CatDistribution)
def _kl_cat_cat(p: "CatDistribution", q: "CatDistribution") -> "Tensor":
    if (p.dim != q.dim) or (len(p.base_dists) != len(q.base_dists)):
        raise NotImplementedError
    return torch.stack(
        [kl_divergence(d, q.base_dists[i]) for i, d in enumerate(p.base_dists)]
    ).sum(dim=0)


@register_kl(CatDistribution, Distribution)
def _kl_cat_distribution(p: "CatDistribution", q: "Distribution") -> "Tensor":
    if len(p.base_dists) > 1:
        raise NotImplementedError
    return kl_divergence(p.base_dists[0], q)


@register_kl(Distribution, CatDistribution)
def _kl_distribution_cat(p: "Distribution", q: "CatDistribution") -> "Tensor":
    if len(q.base_dists) > 1:
        raise NotImplementedError
    return kl_divergence(p, q.base_dists[0])
